"""

    Connection.py
    $Author: payne $
    $Date: 2002/09/03 16:52:07 $

Copyright (C) 2002 Association of Universities for Research in
Astronomy (AURA).  See LICENSE.txt for the usual disclaimers.

    This class defines objects that encapsulate a connection to an
    LDAP directory.  The two objects are the AuthenticatedUser
    object, which is a directory connection where the user has been
    authenticated against the directory, and the AnonymousUser object
    which is an anonymous connection to the directory.  These objects
    contain methods for searching the directory, methods for getting
    particular data from the directory, along with some utilites for
    manipulating DNs and managing roles.  These objects are transient,
    and when used in the Zope layer, one is created for each request.

    The AuthenticatedUser class subclasses the AnonymousUser class.
    The authentication credentials required to create an
    AuthenticatedUser consist of a password and either a DN, or
    a login attribute name (defaults to the iPlanet appropriate 'uid')
    and value.  If the specified person object is not found, then a
    NO_SUCH_OBJECT exception is raised.  If the password is incorrect,
    then an INVALID_CREDENTIAL exception is raised.

    A small MappingObject class provides a convenient way to return
    search results in a format suitable for use with the Zope 
    dtml-in and dtml-tree tags.

"""

__version__ = '$Revision: 1.32 $'[11:-2]
# $Source: /home/payne/job/ldap/CVS/STLDAPManager/Connection.py,v $

from copy import deepcopy
import re
from string import find, join, lower
from urllib import quote_plus

from LDAPExceptions import *
from LDAPUtilities import _ldap, PY_LDAP_VERSION
from LDAPUtilities import *
from SearchLister import SearchLister

class AnonymousUser:
    """
    Encapsulates the actions that any user of the STScI
    Master Directory can do without authenticating
    to the directory.

    """

    def __init__(self, ldapDirectory, zopeUser=None):
        """
        Initializes the object and creates a connection to 
        the LDAP server.  Assigns the user the ANON role.
        If a Zope user object is supplied, its roles are
        also assigned.

        """
        self.distinguishedName = ""
        self.LDAPDirectory = ldapDirectory

        self.bindAnonymously(self.LDAPDirectory.getServer(),
                             self.LDAPDirectory.getPort())

        self._roles = ['ANON',]
        if zopeUser is not None:
            self._roles.extend( list(self._getZopeRoles( zopeUser )) )

    def bindAnonymously(self, server, port):
        """
        Opens a connection to the LDAP server and
        performs an anonymous bind to the directory.
        The connection object is assigned to the
        'Connection' attribute of self.

        """
        # Stolen from Jens Vagepohl's LDAPUserFolder.  Use 'open' with 
        # python-ldap # version 1, but 'initialize' with version 2.
        #
        if PY_LDAP_VERSION[:1] == '2':
            conn_func = _ldap.initialize
            conn_str = '%s://%s:%s' % ( getattr( self, '_conn_proto', 'ldap' ), 
                                        server, 
                                        port )
        else:
            conn_str = '%s:%s' % ( server, port )
            conn_func = _ldap.open

        try:
            self.Connection = apply( conn_func, ( conn_str, ) )
        # Thank you Jens.

        except _ldap.LDAPError, e:
            raise OPEN_FAILURE(source=e,
                               information="server = %s:%s" % ( server, port))
        except Exception, e:
            raise LDAPException(information="Server: %s, Port: %s" % \
                                (server, str(port)))
        try:
            self.Connection.simple_bind_s( '', '' )

        # Also stolen from Jens.  Explicitly set the protocol if there
        # is a complaint about it.
        #
        except _ldap.PROTOCOL_ERROR:
            self.Connection.protocol_version = _ldap.VERSION3
            self.Connection.simple_bind_s( '', '' )

        # Any other problem raises an error
        #
        except _ldap.LDAPError, e:
            raise BIND_FAILURE(source=e)

    def getAllGroups( self, searchBase= ""):
        """
        Returns a list of distinguishedNames of all group-like objects in
        or below the searchBase.

        """
        if searchBase == "":
            searchBase = self.LDAPDirectory.getDirectoryBase()

        query = '(|(objectClass=groupofnames)(objectClass=groupofuniquenames))'
        try:
            groups = self.search(searchBase, _ldap.SCOPE_SUBTREE, query, ['cn'])
            return map( lambda x: x[0], groups )
        except NO_SUCH_OBJECT:
            return []

    def getMatchingGroups( self, groupName, searchBase=""):
        """
        Returns a list of distinguishedNames of all group-like objects 
        in or below the searchBase, whose naming attribute matches
        (case insensitive) groupName.

        """
        return filter( lambda x,y=groupName: lower(getNameFromDN(x))==lower(y),
                       self.getAllGroups( searchBase ) )

    def _getChildren(self, parentDN, query ):
        """
        Returns a list of all of the immediate children
        of parentDN that satisfy the query.

        """
        try:
            list = self.search(parentDN, _ldap.SCOPE_ONELEVEL,
                                 query, ['objectClass'])
            transformedList = map( lambda x: MappingObject(x[0]), list )
            transformedList.sort(self._nameSort)
            return transformedList
        except NO_SUCH_OBJECT:
            return []

    def getChildGroups( self, parentDN ):
        """
        Returns all of the groups that are immediate children
        of the parentDN.

        """
        query = '(|(objectClass=groupofnames)(objectClass=groupofuniquenames))'
        return self._getChildren(parentDN, query)

    def getChildOrgUnits( self, parentDN ):
        """
        Returns all of the organizational units that are
        immediate children of the parentDN.

        """
        query = '(objectClass=organizationalunit)'
        return self._getChildren(parentDN, query)
    

    def getDNfromUserName( self, userName, loginAttr='uid' ):
        """
        Give a userName, returns the distinguishedName for that user.
        If there is no matching user in the directory, raises a
        NO_SUCH_OBJECT exception.

        """
        user = self.search( self.LDAPDirectory.getDirectoryBase(),
                            _ldap.SCOPE_SUBTREE,
                            "%s=%s" % ( loginAttr, userName ) )
        return user[0][0]  

    def getLDAPDirectory( self ):
        """
        Returns the LDAPDirectory singleton.
        """
        return self.LDAPDirectory

    def getNameFromDN( self, distinguishedName ):
        """
        Wrapper for LDAPUtilities.getNameFromDN
        """
        return getNameFromDN( distinguishedName )

    def getReadableSearchKeys(self, schema):
        """
        Returns a filtered list of search keys, containing only
        those readable by the user.
        """
        searchKeys = deepcopy( schema._searchKeys )
        for attributeName in searchKeys:
            attributeSchema = schema.getAttributeSchema(attributeName)
            readers = attributeSchema.getReaders()
            canRead = FALSE
            for role in self.getRoles():
                if role in readers:
                    canRead = TRUE
                    break
            if not canRead:
                searchKeys.remove(attributeName)
        return searchKeys

    def getRoles(self):
        """
        Returns a list of roles that the current user holds.
        """
        return self._roles

    def _getZopeRoles( self, zopeUser ):
        """
        Returns a list of all of the configured roles held by the
        supplied zopeUser object.

        """
        roles = zopeUser.getRoles()
        configuredRoles = \
            self.LDAPDirectory.getManagedRoles().getConfiguredRoleNames()
        return filter( lambda x,y=configuredRoles: x in y, roles )

    def _nameSort( self, x, y ):
        """
        Helper function for sorting MappingObjects
        """
        a = x.mappingName
        b = y.mappingName
        return cmp( a, b )

    def orgUnitMemberBrowser(self,junk,REQUEST,private=0,**args):
        """
        Display org chart search results:
        """
        level = REQUEST['tree-level']
        colspan = REQUEST['tree-colspan']
        dataspan = colspan - level
        if dataspan == 2:
        
            #user = REQUEST['user']
            if REQUEST.has_key('private'):
                private = REQUEST['private']
            else:
                private = 0
            mappingDN = REQUEST['mappingDN']
            schema = self.LDAPDirectory.getPersonSchema()
            firstKey = schema.getSearchKeys()[0]
            namingAttribute = schema.getNamingAttribute()
            if firstKey == namingAttribute:
                searchKeys = [firstKey]
            else:
                searchKeys = [firstKey,namingAttribute]
            try:
                if schema.getPersonsOUAttr() == 'dn':
                    results = SearchLister( self, None, firstKey, 'person',
                                            mappingDN, _ldap.SCOPE_ONELEVEL,
                                            schema.getSearchFilter(),
                                            searchKeys, schema ).resultsList
                else:
                    searchBase = self.LDAPDirectory.getPersonSearchBase()
                    filter = "(&%s(%s=%s))" % ( schema.getSearchFilter(),
                                                schema.getPersonsOUAttr(),
                                                mappingDN )
                    results = SearchLister( self, None, firstKey, 'person',
                                            searchBase, _ldap.SCOPE_SUBTREE,
                                            filter, searchKeys,
                                            schema ).resultsList
            except NO_SUCH_OBJECT:
                return ''

            _td_colspan='<TD COLSPAN="%s" NOWRAP></TD>'
            _td_single ='<TD WIDTH="16" NOWRAP></TD>'
            if level > 1:   h=_td_colspan % (level)
            elif level > 0: h=_td_single  * (level)
            else: h=''

            output = '<TR>%s<TD WIDTH="16" NOWRAP></TD>' % h
            output = output + '<TD%s VALIGN="TOP">' % (
                 level < 2 and (' COLSPAN="%s"' % dataspan) or \
                 ' COLSPAN="%s"' % '1' )
            if private:
                for row in results:
                    output = output + """<a style="text-decoration: none"
href="%s?distinguishedName=%s">
&nbsp;&nbsp;%s</a><br>""" % ( REQUEST['URL0'],
                              quote_plus(row['distinguishedName'][0]),
                              row[firstKey][0] )
            else:
                for row in results:
                    output = output + """<a style="text-decoration: none"
href="%s?query=%s%%3D%s">
&nbsp;&nbsp;%s %s %s</a><br>""" % ( REQUEST['URL0'],
                              namingAttribute,
                              row[namingAttribute][0],
                              row[firstKey][0],
                              '', '' )
                              #colspan, level )
            output = output + '</TD></TR>'
            return output
        else:
            return ''

    def search(self, base, scope, query, attrList = None):
        """
        Performs a generic search of the LDAP directory and returns the
        result.  If the search does not find anything, raises a
        NO_SUCH_OBJECT exception.

        Returns a sequence of results.  The sequence items are themselves
        a sequence containing the items distinguished name and a dictionary
        of attribute name/value pairs.  Each value item is itself a sequence
        of strings.
        """
        try:
            result = self.Connection.search_s(base, scope, query, attrList)
        except _ldap.LDAPError, e:
            raise NO_SUCH_OBJECT(source=e,
                                 information='Searchbase: %s, Query: %s' %(base, query))
        if result:
            return result
        else:
            raise NO_SUCH_OBJECT(information='Searchbase: %s, Query: %s' %(base, query))

    def showSearchResults(self, objectQuery, sortKey, objectCategory=None,
                          searchBase=None, searchScope=None, filter=None, 
                          searchKeys=None ):
        """
        Returns a list of objects selected by the query.

        """
        if objectCategory is None:
            schema = self.schema
            objectCategory = self.objectCategory
        else:
            if lower(objectCategory) == 'person':
                schema = self.LDAPDirectory.getPersonSchema()
            elif lower(objectCategory) == 'orgunit':
                schema = self.LDAPDirectory.getOrgUnitSchema()
            elif lower(objectCategory) == 'group':
                schema = self.LDAPDirectory.getGroupSchema()
            else:
                raise UNRECOGNIZED_OBJECT

        return SearchLister( self, objectQuery, sortKey, objectCategory, 
                             searchBase, searchScope, filter, searchKeys, 
                             schema ).resultsList

    def unbind(self):
        """
        Closes the connection to the directory.
        """
        try:
            self.Connection.unbind()
        except _ldap.LDAPError, e:
            raise LDAPException(source=e, information='Unbind failed.')

    def whatKindOfGroup(self,groupDN):
        """
        Given a group DN, return 'groupOfNames' or 'groupOfUniqueNames',
        depending on which it is, or None if the answer is inaccessible
        or neither.
        """
        try:
            objectClasses = self.search(groupDN,
                                        _ldap.SCOPE_BASE, 
                                        "(objectclass=*)", 
                                        ['objectClass'])[0][1]['objectClass']
            if 'groupofnames' in objectClasses:
                return 'groupOfNames'
            elif 'groupofuniquenames' in objectClasses:
                return 'groupOfUniqueNames'
            else:
                return None
        except:
            return None

class AuthenticatedUser(AnonymousUser):
    """
    Encasulates the actions that a user can perform after
    authenticating to the STScI Master Directory in addition
    to those actions that an anonymous user can perform.

    """

    def __init__(self, userName, passwd, ldapDirectory, loginAttr='uid',
                 zopeUser=None):
        """
        Attempts to authenticate to the directory with the
        supplied credentials.  If that authentication succeeds,
        determines which roles this user holds and sets the
        other attributes of the authenticated user.

        """

        AnonymousUser.__init__(self, ldapDirectory, zopeUser=zopeUser)

        # Perhaps 'userName' is a complete valid DN.  See if a search on it
        # works.  If not, try getDNfromUserName. 

        try:
            user = self.search( userName,
                                _ldap.SCOPE_BASE, 
                                "(objectclass=*)" )
            self.distinguishedName = userName
        except:
            try:
                distinguishedName = self.getDNfromUserName(userName, 
                                                       loginAttr=loginAttr)
                self.distinguishedName = distinguishedName
            except:
                raise NO_SUCH_OBJECT

        # Now try connecting

        try:
            self.Connection.simple_bind_s(self.distinguishedName, passwd)

        except _ldap.LDAPError, e:
            raise INVALID_CREDENTIALS(source = e,
                                      information = 'Username: %s' % userName )

        user = self.search( self.distinguishedName, 
                            _ldap.SCOPE_BASE, 
                            "(objectclass=*)" )
        self.attributes = user[0][1]
        
        self._roles.append( 'AUTH' )
        self._roles.extend( self._getLDAPGroupRoles() )

    def addRole(self, role):
        """
        Adds the given role to the user's list of roles.
        """
        if not role in self._roles:
            self._roles.append(role)
        
    def attribute(self, name):
        """
        Returns sequence containing the values of the user's
        attribute named 'name'.  Returns None if that attribute
        either is invalid or is not populated.
        """
        if self.attributes.has_key(name):
            return self.attributes[name]
        else:
            return None

    def _getLDAPGroupRoles(self):
        """
        Returns a list of all of the configured roles based on LDAP 
        groups that this user is a member of.
        """
        configuredRoles = \
            self.LDAPDirectory.getManagedRoles().getConfiguredRoleNames()

        query = \
            "(|(&(objectClass=groupOfNames)(member=%s))(&(objectClass=groupOfUniqueNames)(uniqueMember=%s)))" \
              % (self.distinguishedName, self.distinguishedName)
        try:
            roles = self.search(self.LDAPDirectory.getZopeRoleSearchBase(),
                                self.LDAPDirectory.getZopeRoleSearchScope(),
                                query,
                                ['cn'])
            return filter( lambda x,y=configuredRoles: x in y,
                           map( lambda x: getNameFromDN(x[0]), roles ))
        except NO_SUCH_OBJECT:
            return []

    def rename(self, currentDN, newRDN ):
        """
        Renames an object (or, more correctly, changes an object's relative
        distinguished name).

        """
        try:
            self.Connection.modrdn_s(currentDN, newRDN)
        except _ldap.LDAPError, e:
            raise RENAME_ERROR(source = e,
                               information = currentDN )

class MappingObject:
    """
    This object is defined as a convenience.  By having getChildGroups and
    getChildOrgUnits return a list of MappingObjects instead of a list of
    dictionaries, it makes it easier to use the Zope tree tag.

    """
    def __init__(self, mappingDN):
        self.mappingDN = mappingDN
        self.mappingName = getNameFromDN(mappingDN)

    def __repr__(self):
        return '[' + self.mappingDN + ': ' + self.mappingName + ']'
