"""

    LDAPObject.py
    $Author: payne $
    $Date: 2002/01/03 20:21:12 

Copyright (C) 2002 Association of Universities for Research in
Astronomy (AURA).  See LICENSE.txt for the usual disclaimers.

    The LDAPObject class represents an object in the LDAP directory.
    It contains methods for creating and deleting objects, and for
    manipulating and reporting information about the attributes of
    the object.  Objects are created as a shell containing only a
    Connection object.  

    The 'refresh' method, given a DN, will retrieve the attributes of
    that object in the directory, and use them to decide what kind
    of object it is ('person', 'orgUnit', or 'group'), and to build
    a dictionary of ManagedAttribute objects, each of which contains
    the schema and values for a single attribute.

    The 'updateDirectory' method finds all of the ManagedAttribute
    objects whose values have been changed, and writes those values
    back to the directory.

    In the Zope layer, LDAPObjects are transient, instantiated as an
    attribute of a ZObject created to handle a single request.  The
    LDAPObject is refreshed from a particular DN, its attributes are
    changed in response to forms submitted by the user, and then those
    changes are written back to the directory.

"""

__version__ = '$Revision: 1.35 $'[11:-2]

from string import find, lower

from LDAPExceptions import *
from LDAPUtilities import _ldap, DNNormalize, DNsAreEquivalent, \
    getNameFromDN, listsAreEquivalent, listCopy, list1ContainsList2, \
    mergeLists, TRUE, FALSE
from ManagedAttributeSchema import ManagedAttribute


class LDAPObject:
    
    def __init__( self, user ):
        """
        Save the user object, and supply shortcuts for accessing
        certain user attributes.
        """

        self.user = user
        self.Connection = user.Connection
        self.searchBase = user.LDAPDirectory.getDirectoryBase()
        self.personSearchBase = user.LDAPDirectory.getPersonSearchBase()
        self.orgUnitBrowseBase = user.LDAPDirectory.getOrgUnitBrowseBase()
        self.groupBrowseBase = user.LDAPDirectory.getGroupBrowseBase()
        self.managedRoles = user.LDAPDirectory.getManagedRoles()
        self.distinguishedName = ""
        
    def canReadAttribute(self, attributeName):
        """
        Returns true if the current user has read access to
        the named attribute based on the roles the user holds.
        """
        readers = self._ManagedAttributes[attributeName].getReaders()
        for role in self.user.getRoles():
            if role in readers:
                return TRUE
        return FALSE
 
    def canWriteAttribute(self, attributeName):
        """
        Returns true if the current user has write access to
        the named attribute based on the roles the user holds.
        """
        writers = self._ManagedAttributes[attributeName].getWriters()
        for role in self.user.getRoles():
            if role in writers:
                return TRUE
        return FALSE

    def create(self, distinguishedName, objectCategory = 'None'):
        """
        Creates a new object of the appropriate type.  Adds defaults
        for all required attributes except objectClass, the naming attribute,
        and attributes which have required values.
        """
        objectName = getNameFromDN(distinguishedName)

        if lower(objectCategory) == 'group':
            self.managedSchema = self.user.LDAPDirectory.getGroupSchema()
        elif lower(objectCategory) == 'orgunit':
            self.managedSchema = self.user.LDAPDirectory.getOrgUnitSchema()
        elif lower(objectCategory) == 'person':
            self.managedSchema = self.user.LDAPDirectory.getPersonSchema()
        else:
            raise UNRECOGNIZED_OBJECT
    
        attrList = []

        for attrName in self.managedSchema.getRequiredAttributeNames():
            if attrName == 'objectClass':
                attrList.append(('objectClass',self.getRequiredObjectClasses()))
            elif attrName == self.managedSchema.getNamingAttribute():
                attrList.append( (attrName, [objectName]) )
            elif attrName == 'member' or attrName == 'uniqueMember':
                # since OpenLDAP requires these attributes on a 
                # groupofname or groupofuniquenames, respectively
                attr = self.managedSchema.getAttributeSchema(attrName)
                values = attr.getRequiredValues()
                if values == []:
                    values = ['cn=Default Value']
                attrList.append( (attrName, values) )
            else:
                attr = self.managedSchema.getAttributeSchema(attrName)
                values = attr.getRequiredValues()
                if values == []:
                    values = ['Default Value']
                attrList.append( (attrName, values) )
                
        try:
            self.Connection.add_s(distinguishedName, attrList)
        except _ldap.ALREADY_EXISTS:
            raise ALREADY_EXISTS
#        except _ldap.OBJECT_CLASS_VIOLATION:
#            raise "Problem with attrList='%s'" % attrList

        self.refresh(distinguishedName)
        return objectName
        
    def delete( self, areYouSure = FALSE ):
        """
        Deletes the object object from the LDAP directory
        and invalidated the corresponding Python instance.
        """
        if areYouSure:
            self._deleteObject( self.distinguishedName )
            self.distinguishedName = None
            self.allowedAttributes = None


    def _deleteObject( self, distinguishedName ):
        """
        Recursively deletes all of the children of the current
        object and the deletes the object.
        """
        try:
            children = self.user.search( distinguishedName,
                                         _ldap.SCOPE_ONELEVEL,
                                         '(objectClass=*)',
                                         ['objectClass'] )
            for child in children:
                self._deleteObject( child[0] )
        except NO_SUCH_OBJECT:
            pass

        #   Before we delete this object, we should scan the
        #   entire directory looking for objects that have
        #   attributes of syntax 'DN' and that contain entries
        #   that point to this object
        
        try:
            self.Connection.delete_s( distinguishedName )
        except _ldap.LDAPError, e:
            raise DELETE_FAILURE(source=e,
                information = 'DN = %s' % (self.distinguishedName))

    def displayCategoryHasReadableAttributes( self, displayCategory ):
        """
        Returns TRUE or FALSE depending on whether the given
        display category contains any items the user can read.
        """
        for aName in self._ManagedAttributes.keys():
            categories = self._ManagedAttributes[aName].getDisplayCategories()
            if self._ManagedAttributes[aName].getManaged() \
             and displayCategory in categories \
             and self.canReadAttribute(aName):
                return TRUE
        return FALSE

    def displayCategoryHasWritableAttributes( self, displayCategory ):
        """
        Returns TRUE or FALSE depending on whether the given
        display category contains any items the user can write.
        """
        for aName in self._ManagedAttributes.keys():
            categories = self._ManagedAttributes[aName].getDisplayCategories()
            if self._ManagedAttributes[aName].getManaged() \
             and displayCategory in categories \
             and self.canWriteAttribute(aName):
                return TRUE
        return FALSE

    def friendlyNameSort( self, x, y ):
        """
        Helper function for getAttributes.
        """
        x = self._ManagedAttributes[x].getFriendlyName( )
        y = self._ManagedAttributes[y].getFriendlyName( )
        return cmp( x, y )

    def getNameFromDN( self, distinguishedName, REQUEST=None ):
        """
        Wrapper for LDAPUtilities.getNameFromDN
        """
        return getNameFromDN( distinguishedName )

    def getRequiredObjectClasses(self):
        """
        Returns a list of objectClasses that the schema requires
        be in the objectClass attribute for this object.
        """
        oClassSchema = self.managedSchema.getAttributeSchema('objectClass')
        return oClassSchema.getRequiredValues()

    def getManagedAttribute( self, attributeName ):
        """
        Returns the  ManagedAttribute object for the named attribute.
        """
        if not self._ManagedAttributes.has_key(attributeName):
            raise NO_SUCH_ATTRIBUTE(information = \
                                    'Attribute Name = %s' % \
                                    (attributeName))
        elif not self._ManagedAttributes[attributeName].getManaged():
            raise INSUFFICIENT_ACCESS(information = \
                                    'Attribute Name = %s' % \
                                    (attributeName))
        elif not self.canReadAttribute(attributeName):
            raise INSUFFICIENT_ACCESS(information = \
                                    'Attribute Name = %s' % \
                                    (attributeName))
        else:
            return self._ManagedAttributes[attributeName]
        

    def getManagedAttributeNames(self, displayCategory):
        """
        Returns a list of managed attribute names in the given
        display category and for which the user has at least
        read access.
        """
        names = filter( lambda aName,s=self,dc=displayCategory: 
                  s._ManagedAttributes[aName].getManaged() 
                  and dc in s._ManagedAttributes[aName].getDisplayCategories() 
                  and s.canReadAttribute(aName), self._ManagedAttributes.keys())
        names.sort(self.friendlyNameSort)
        return names

    def getObjectCategory(self):
        """
        Returns one of 'group', 'orgunit', 'person', defining the type of
        this object, or 'None' if the object has not been refreshed.
        """
        if hasattr( self, 'distinguishedName' ):
            if hasattr( self, '_objectCategory' ):
                return self._objectCategory
        else:
            return 'None'

    def refresh( self, distinguishedName = None ):
        """
        Reload the allowed attributes for the current
        instance from the underlying LDAP directory.
        """

        if self.distinguishedName != '':
            distinguishedName = self.distinguishedName

        if distinguishedName is None:
            raise NOT_ENOUGH_INFORMATION

        retrievedObject = self.user.search(distinguishedName,
                                           _ldap.SCOPE_BASE,
                                           '(objectclass=*)')

        self._dirty = FALSE
        self.distinguishedName = retrievedObject[0][0]
        firstComma = find(self.distinguishedName,',')
        self.parentDN = self.distinguishedName[firstComma+1:]

        #   Really seeing that the user is not an AnonymousUser:

        if self.user.distinguishedName != "":
            if DNsAreEquivalent( self.distinguishedName, 
                                 self.user.distinguishedName ):
                self.user.addRole('SELF')
            
        retrievedAttributeValues = retrievedObject[0][1]

        oClass = map( lower, retrievedAttributeValues['objectClass'] )

        if 'organizationalunit' in oClass:
            self.managedSchema = self.user.LDAPDirectory.getOrgUnitSchema()
            self._objectCategory = 'orgunit'
        elif 'person'in oClass:
            self.managedSchema = self.user.LDAPDirectory.getPersonSchema()
            self._objectCategory = 'person'
        elif ('groupofnames' in oClass) or ('groupofuniquenames' in oClass):
            self.managedSchema = self.user.LDAPDirectory.getGroupSchema()
            self._objectCategory = 'group'
        else:
            raise UNRECOGNIZED_OBJECT

        self._ManagedAttributes = {}
        
        getSchema = self.managedSchema.getAttributeSchema
        for attributeName in self.managedSchema.getAllowedAttributeNames():
            
            metadata = getSchema(attributeName).getMetadata()

            values = retrievedAttributeValues.get(attributeName, [])
            attribute = ManagedAttribute(attributeName, values, metadata)
            self._ManagedAttributes[attributeName] = attribute

        if self._ManagedAttributes.has_key('manager'):
            mgrList =  self._ManagedAttributes['manager'].getValues()
            for mgr in mgrList:
                if DNsAreEquivalent(mgr, self.user.distinguishedName):
                    self.user.addRole('MNGR')
                    break

        return self.distinguishedName

    def rename( self, newObjectName ):
        """
        Renames the current object by changing it's rdn.
        """
        firstComma = find(self.distinguishedName,',')
        self.parentDN = self.distinguishedName[firstComma+1:]
        newRDN = self.managedSchema.getNamingAttribute() + '=' + newObjectName
        try:
            self.Connection.modrdn_s( self.distinguishedName, newRDN )
        except _ldap.ALREADY_EXISTS:
            raise ALREADY_EXISTS
        
        self.distinguishedName = newRDN + ',' + self.parentDN
        return self.refresh()

    def setManagedAttribute( self, managedAttribute ):
        """
        Returns the ManagedAttribute object for the named attribute.
        """
        attributeName = managedAttribute.getAttributeName()
        
        if not self._ManagedAttributes.has_key(attributeName):
            raise NO_SUCH_ATTRIBUTE(information = \
                                    'Attribute Name = %s' % \
                                    (attributeName))
        elif not self._ManagedAttributes[attributeName].getManaged():
            raise INSUFFICIENT_ACCESS(information = \
                                    'Attribute Name = %s' % \
                                    (attributeName))
        elif not self.canWriteAttribute(attributeName):
            raise INSUFFICIENT_ACCESS(information = \
                                    'Attribute Name = %s' % \
                                    (attributeName))
        else:
            self._ManagedAttributes[attributeName] = managedAttribute
            if managedAttribute.ValueChanged():
                self._dirty = TRUE
        
    def updateDirectory( self ):
        """
        Writes any changes that have been made to this object to
        the LDAP directory.
        """
        changedAttrNames = filter( lambda aName,s=self: 
                           s._ManagedAttributes[aName].getManaged() 
                           and s._ManagedAttributes[aName].ValueChanged()
                           and s.canReadAttribute(aName), 
                           self._ManagedAttributes.keys())

        ldapModList = []

        for attrName in changedAttrNames:
            attrValues = self._ManagedAttributes[attrName].getValues()
            if attrValues == []:
                mode = _ldap.MOD_DELETE
            else:
                mode = _ldap.MOD_REPLACE
            ldapModList.append( ( mode, attrName, attrValues ) )

        if ldapModList != []:
            try:
                self.Connection.modify_s( self.distinguishedName, ldapModList )
            except _ldap.INSUFFICIENT_ACCESS:
                raise INSUFFICIENT_ACCESS
            except _ldap.NO_SUCH_OBJECT:
                raise NO_SUCH_OBJECT
            
        self._dirty = FALSE
