"""
    The LDAPSchema class.

    LDAPSchema.py
    $Author: payne $
    $Date: 2002/06/14 01:45:06 $

Copyright (C) 2002 Association of Universities for Research in
Astronomy (AURA).  See LICENSE.txt for the usual disclaimers.

    This class encapsulates an LDAP directory schema for a
    particular category of object ('person', 'orgUnit', or 'group').
    A singleton for each category persists as attributes of an
    LDAPDirectory object.  

    An LDAPSchema contains information obtained from the directory,
    like the list of all of the attributes allowed for this category
    of object, saving the trouble of retrieving it for every 
    subsequent request.  It also contains configuration choices, 
    such as the subset of allowed attributes that will actually
    be used and managed.  And finally, it contains information that
    is in the directory, but not available, such a the list of 
    attributes that a particular user is able to change.  For these,
    the idea is either to configure this object to match the directory,
    or to configure this object with the access rules that you want,
    and then to extract that information in a form suitable for
    re-configuring the directory.

    An LDAPSchema consists primarily of a dictionary of
    ManagedAttributeSchema objects.  It is instantiated by passing
    it a list of object classes that corresponds to the object
    category at hand (e.g., ['top','person','organizationalPerson',
    'inetOrgPerson'] for the 'person' object category).  This
    initial list of objects is found in the SchemaDefaults module.
    The LDIF representation of the schema is retrieved from the
    directory and parsed into lists of required and optional
    attributes defined in the schemas for the object class list.
    These lists are then used to build the dictionary.

    In the Zope layer, a transient ZSchema object, wrapping an
    LDAPObject, is used for updating the schema.

    Just a reminder:  'allowed' attributes are all of the 
    attributes defined in the directory for the given 
    object classes or category, 'required' attributes are those
    required in the schema, and 'optional' attributes are those
    that are allowed but not required.

"""

__version__ = '$Revision: 1.17 $'[11:-1]

from string import lower

from LDAPExceptions import *
from LDAPSyntax import getNameFromOID
from LDAPUtilities import _ldap, TRUE, FALSE, listsAreEquivalent
from LDIFObjectParser import LDIFObjectParser
from LDIFAttributeParser import LDIFAttributeParser
from ManagedAttributeSchema import ManagedAttributeSchema

class LDAPSchema:
    """
        A class to encapsulate an LDAP schema, obtained from the
        LDAP directory.
    """

    def __init__( self, user, objectClasses ):

        directoryBase = user.LDAPDirectory.getDirectoryBase()

        schemaDN = user.search( '', 
            _ldap.SCOPE_BASE, 
            "(objectclass=*)", 
            ['subschemaSubentry',])[0][1]['subschemaSubentry'][0]

        schema = user.search( schemaDN, _ldap.SCOPE_BASE,
            "(objectclass=*)", ['objectClasses','attributeTypes'] )[0][1]

        self._objectSchemas = \
                self._parseLDIFObjectDefinitions(schema['objectClasses'])
        self._attributeSchemas = \
                self._parseLIDFAttrDefinitions(schema['attributeTypes'])

        self.requiredAttributeNames, self.optionalAttributeNames \
                                     = self._buildAttributeLists(objectClasses)

        self.allowedAttributeNames = self.requiredAttributeNames \
                                   + self.optionalAttributeNames

        self._attributeCollection = \
            self._buildManagedAttributeCollection( objectClasses )

    def _buildAttributeLists(self, objectClasses):
        """
            Returns lists of required and optional attribute names
            for the composite LDAP object made up of all of the
            given objectClasses
        """
        optionalAttributeNames = []
        requiredAttributeNames = []

        if type(objectClasses) == type(''):
            objectClasses = [objectClasses]
            
        for objectClass in objectClasses:
            required, optional = self._getAttributeNames(objectClass)
            requiredAttributeNames = requiredAttributeNames + \
                filter(lambda i,r=requiredAttributeNames: i not in r, required)
            optionalAttributeNames = optionalAttributeNames + \
                filter(lambda i,o=optionalAttributeNames: i not in o, optional)
                                           
        optionalAttributeNames = filter(
                                 lambda i,r=requiredAttributeNames: i not in r,
                                 optionalAttributeNames )

        return requiredAttributeNames, optionalAttributeNames

    def _buildManagedAttributeCollection( self, objectClasses ):
        """
            Returns a dictionary, keyed to attribute name and
            containing ManagedAttributeSchema objects for each
            allowed attribute for the objects in the objectClasses
            list.
        """
        attributeCollection = {}

        for attributeName in self.allowedAttributeNames:
            parsedAttribute = self._attributeSchemas[lower(attributeName)]
            metadata = {}
            metadata['syntax'] = parsedAttribute.syntax
            if attributeName in self.requiredAttributeNames:
                metadata['required'] = TRUE
            managedAttribute = ManagedAttributeSchema( attributeName, metadata )
            attributeCollection[ attributeName ] = managedAttribute

        return attributeCollection
    
    def _getAttributeNames(self, objectClass):
        """
            Returns lists of required and optional attributes for the
            given LDAP object class.
        """
        LDIFObject = self._objectSchemas[lower(objectClass)]
        required = LDIFObject.mustContain
        optional = LDIFObject.mayContain

        if LDIFObject.superClass:
            supRequired, supOptional = \
                    self._getAttributeNames(LDIFObject.superClass)
            required = required + filter( lambda i,r=required: i not in r,
                                          supRequired )
            optional = optional + filter( lambda i,o=optional: i not in o,
                                          supOptional )

        return required, optional            

    def getManagedAttributeCollection(self):
        """
            Returns a dictionary of ManagedAttributeSchema objects.
        """
        return self._attributeCollection.copy()

    def getManagedAttributeSchema( self, ldapName ):
        """ 
            Returns a named ManagedAttributeSchema object.
        """
        assert ldapName in self._attributeCollection.keys()
        return self._attributeCollection[ldapName]

    def _parseLIDFAttrDefinitions(self, LDIFExpressionList):
        """
            Parses a list of LDIF attribute definitions and returns a
            dictionary of LDIFAttributeParser objects.
        """
        attributeSchemas = {}
        for ldifExpression in LDIFExpressionList:
            parsedObject = LDIFAttributeParser( ldifExpression )
            attributeSchemas[lower(parsedObject.name)] = parsedObject
        return attributeSchemas
                   
    def _parseLDIFObjectDefinitions(self, LDIFExpressionList):
        """
            Parses a list of LDIF object definitions and returns a
            dictionary of LDIFObjectParser objects.
        """
        objectSchemas = {}
        for ldifExpression in LDIFExpressionList:
            parsedObject = LDIFObjectParser( ldifExpression )
            objectSchemas[lower(parsedObject.name)] = parsedObject
        return objectSchemas

    
