"""
    LDIFAttributeParser.py
    $Author: payne $
    $Date: 2002/06/18 20:05:01 $

Copyright (C) 2002 Association of Universities for Research in
Astronomy (AURA).  See LICENSE.txt for the usual disclaimers.

    The LDIFAttributeParser provides methods for parsing an LDIF 
    attribute definition, and converting the result back to LDIF.

    The general form of an LDIF attribute definition is as follows:
 
    ( oid NAME 'attributeName' DESC 'attribute description'
      EQUALITY matching-rule SYNTAX oid SINGLE-VALUE
      X-ORIGIN 'who defined this' )
 
    or, in OpenLDAP
 
    ( oid NAME 'attributeName' SUP previouslyDefinedAttributeName )
 
    Each field is as follows:
 
        oid  - required.  A string of the form n.m.x.y.z...
        NAME - required.  Always surrounded with single quotes.  May be
                          a list of aliases.
        DESC - optional.  If it is given, the value is surrounded
                          by single quotes.
        EQUALITY - optional.  Defines the rule for equality for
                          this 
        SYNTAX - required unless SUP is present. OID of the syntax 
                          definition that this attribute follows.
        SINGLE-VALUE - optional.  If present, this attribute can
                          have only a single value; otherwise it
                          may have multiple values.
        X-ORIGIN - optional.  If given, the value is surrounded
                          in single quotes.
        SUP               The superior attribute from which attributes
                          like SYNTAX and EQUALITY are inherited

"""

__version__ = '$Revision: 1.9 $'[11:-1]

import re
from string import find, replace, split

from LDAPUtilities import TRUE, FALSE

whitespacePat = re.compile(r'\s+')

class LDIFAttributeParser:

    def __init__( self, LDIFExpression = None):

        if LDIFExpression:

            self._parseLDIF( LDIFExpression )

    def getLDIF(self):
        """
            Returns the LDIF string defining the objectClass.
        """
        output = '( ' + self.oid

        if self.aliases:
            name = " NAME ( '" + self.name + "'"
            for alias in self.aliases:
                name = name + " '" + alias + "'"
            name = name + " )"
        else:
            name = " NAME '" + self.name + "'"
        output = output + name

        if self.description:
            output = output + " DESC '" + self.description + "'"

        if self.matchingRule:
            output = output + " EQUALITY " + self.matchingRule 

        if self.syntax:
            output = output + " SYNTAX " + self.syntax

        if self.singleValued:
            output = output + " SINGLE-VALUE"

        if self.origin:
            output = output + " X-ORIGIN '" + self.origin + "'"

        if self.sup:
            output = output + " SUP " + self.sup

        output = output + ' )'

        return output

    def getMetadata(self):
        """
        Return the attributes as a dictionary
        """
        dict = {}
        
        for item in ['syntax','sup']:
            dict[item] = getattr( self, item )
        return dict
        
    def inheritsFrom( self, superiorAttr ):
        """
        Inherit values from a superior attribute
        """
        self.syntax = superiorAttr.syntax
        self.singleValued = superiorAttr.singleValued
        
    def _parseLDIF( self, LDIFExpression ):

        # Replace one or more occurences of whitespace with a single space

        LDIFExpression = re.sub( whitespacePat, " ", LDIFExpression )

        # Strip surrounding parentheses if they are there.

        if LDIFExpression[0:2] == '( ':
            ldif = LDIFExpression[2:-2]
        else:
            ldif = LDIFExpression

        i = find(ldif,' ')
        self.oid = ldif[0:i]

        self.aliases = self._getAliases(ldif)
        self.name = self.aliases[0]
        self.aliases.remove(self.aliases[0])
        
        i = find(ldif,"DESC '")
        if i != -1:
            i = find(ldif,"'", i)
            j = find(ldif,"'", i + 1)
            self.description = ldif[i+1:j]
        else:
            self.description = ""

        i = find(ldif,"EQUALITY ")
        if i != -1:
            i = find(ldif," ", i)
            j = find(ldif," ", i + 1)
            if j == -1:
                self.matchingRule = ldif[i+1:]
            else:
                self.matchingRule = ldif[i+1:j]
        else:
            self.matchingRule = ""

        i = find(ldif,"SYNTAX ")
        if i != -1:
            i = find(ldif," ", i)
            j = find(ldif," ", i + 1)
            if j == -1:
                self.syntax = ldif[i+1:]
            else:
                self.syntax = ldif[i+1:j]
        else:
            self.syntax = ""

        if find(ldif," SINGLE-VALUE " ) != -1:
            self.singleValued = TRUE
        else:
            self.singleValued = FALSE
        
        i = find(ldif," X-ORIGIN " )
        if i != -1:
            i = find(ldif, "'", i)
            j = find(ldif, "'", i+1)
            if j == -1:
                self.origin = ldif[i+1 : ]
            else:
                self.origin = ldif[i+1 : j]
        else:
            self.origin = ''

        i = find(ldif,"SUP ")
        if i != -1:
            i = find(ldif," ", i)
            j = find(ldif," ", i + 1)
            if j == -1:
                self.sup = ldif[i+1:]
            else:
                self.sup = ldif[i+1:j]
        else:
            self.sup = ""

    def _getAliases( self, ldif):
        
        i = find(ldif,"NAME")
        if i == -1:
            return []
        i = find(ldif," ", i+1)
        if ldif[i+1:i+2] == "(":
            i = i + 2
            j = find(ldif, " )", i)
        else:
            j = find(ldif, " ", i+1 )
        tmp = replace( ldif[i+1 : j], "'", "")
        return split(tmp, " ")
