"""

    Module: $RCSfile: TestAuthenticatedUser.py,v $
    Author: $Author: payne $
    Date:   $Date: 2002/06/25 22:09:37 $
    
Copyright (C) 2002 Association of Universities for Research in
Astronomy (AURA).  See LICENSE.txt for the usual disclaimers.

"""
__version__ = '$Revision: 1.14 $'[11:-2]

import sys
sys.path.insert(0,'..')

from Connection import *
from LDAPUtilities import _ldap
from LDAPUtilities import *
from TestDirectory import getTestDirectory
import unittest

class testAuthenticatedUser(unittest.TestCase):
    """ Tests the Authenticated User class in the Connection module """

    def setUp(self):
        self.directory = getTestDirectory()
        self.user = AuthenticatedUser('uid=testadmin,ou=people,o=test', 
                                      'testadmin', self.directory)

    def test01(self):
        """ 01 testAuthenticatedUser: Login with DN test """
        self.failUnless(self.user.distinguishedName)

    def test02(self):
        """ 02 testAuthenticatedUser: Login with UID test """
        try:
            user = AuthenticatedUser('testadmin', 
                                     'testadmin',
                                     self.directory,
                                     loginAttr='uid' )
        except:
            #self.fail()
            raise

    def test03(self):
        """ 03 testAuthenticatedUser: Login with CN test """
        try:
            user = AuthenticatedUser('test admin', 
                                     'testadmin',
                                     self.directory,
                                     loginAttr='cn' )
        except:
            self.fail()

    def test04(self):
        """ 04 testAuthenticatedUser: Login with valid DN, invalid passwd """
        self.assertRaises(INVALID_CREDENTIALS,
                          AuthenticatedUser,
                          'uid=testadmin,ou=people,o=test',
                          'wrongPassword',
                          self.directory )

    def test05(self):
        """ 05 testAuthenticatedUser: Login with valid UID, invalid passwd"""
        self.assertRaises(INVALID_CREDENTIALS,
                          AuthenticatedUser,
                          'testadmin', 
                          'wrongPassword',
                          self.directory,
                          loginAttr='uid' )

    def test06(self):
        """ 06 testAuthenticatedUser: Login with valid CN, invalid passwd"""
        self.assertRaises(INVALID_CREDENTIALS,
                          AuthenticatedUser,
                          'test admin', 
                          'wrongPassword',
                          self.directory,
                          loginAttr='cn' )

    def test07(self):
        """ 07 testAuthenticatedUser: Login with invalid DN """
        self.assertRaises(NO_SUCH_OBJECT, 
                          AuthenticatedUser, 
                          'uid=noname,ou=people,o=test', 
                          'wrongpassword',
                          self.directory, )

    def test08(self):
        """ 08 testAuthenticatedUser: Login with invalid UID """
        self.assertRaises(NO_SUCH_OBJECT,
                          AuthenticatedUser, 
                          'noname', 
                          'wrongpassword',
                          self.directory, 
                          loginAttr='uid')

    def test09(self):
        """ 09 testAuthenticatedUser: Specific Attribute Test """
        self.failUnless(self.user.attribute('cn'))

    def test10(self):
        """ 10 testAuthenticatedUser: Bad Specific Attribute Test """
        self.failIf(self.user.attribute('badAttribute'))

    def test11(self):
        """ 11 testAuthenticatedUser: Rename real user test """
        self.user.rename( 'uid=usera,ou=People,o=test', 'uid=bogusUser')
        self.user.rename( 'uid=bogusUser,ou=People,o=test', 'uid=usera' )

    def test12(self):
        """ 12 testAuthenticatedUser: Rename user with collision """
        self.assertRaises(RENAME_ERROR,
                          self.user.rename,
                          'uid=usera,ou=People,o=test',
                          'uid=userb' )

    def test13(self):
        """ 13 testAuthenticatedUser: testing getAllGroups """
        testGroups = ['Directory Administrators',
                      'Divisional Administrators',
                      'Empty Group',
                      'Manager',
                      'System Administrators',
                      'Test Group']
        groups = self.user.getAllGroups('ou=Groups,o=test')
        groups = [getNameFromDN(item) for item in groups]
        self.failUnless(listsAreEquivalent(groups, testGroups))

    def test14(self):
        """ 14 testAuthenticatedUser: testing _getZopeRoles """
        roles = ['ANON',
                 'AUTH',
                 'System Administrators',
                 'Divisional Administrators',]
        self.failUnless( listsAreEquivalent(roles, self.user.getRoles()))

    def test15(self):
        """ 15 testAuthenticatedUser: testing Direct Manager access """
        # usera is manager to USERB
        usera = AuthenticatedUser('uid=usera,ou=people,o=test','usera',
                                   self.directory)
        result = usera.search('uid=USERB,ou=people,o=test',
                              _ldap.SCOPE_BASE,
                              '(objectclass=*)',
                              ['homePhone'])[0][1]
        self.failUnless( result['homePhone'][0] == '123-456-7890' )
        userb = AuthenticatedUser('uid=userb,ou=people,o=test','userb',
                                   self.directory)
        result = userb.search('uid=usera,ou=people,o=test',
                              _ldap.SCOPE_BASE,
                              '(objectclass=*)',
                              ['homePhone'])[0][1]
        self.failIf( result.has_key('homePhone') )

    def test16(self):
        """ 16 testAuthenticatedUser: testing group access """
        # testadmin is in 'Divisional Administrators', so can read
        # homePhone.
        usera = AuthenticatedUser('uid=testadmin,ou=people,o=test','testadmin',
                                   self.directory)
        result = usera.search('uid=USERB,ou=people,o=test',
                              _ldap.SCOPE_BASE,
                              '(objectclass=*)',
                              ['homePhone'])[0][1]
        self.failUnless( result['homePhone'][0] == '123-456-7890' )
        userb = AuthenticatedUser('uid=userb,ou=people,o=test','userb',
                                   self.directory)
        result = userb.search('uid=usera,ou=people,o=test',
                              _ldap.SCOPE_BASE,
                              '(objectclass=*)',
                              ['homePhone'])[0][1]
        self.failIf( result.has_key('homePhone') )

    def tearDown(self):
        self.user.unbind()

if __name__ == '__main__':
    
    tests = unittest.makeSuite(testAuthenticatedUser, 'test')
    unittest.TextTestRunner(verbosity=2).run(tests)
